@@ -1,6 +1,6 @@
 ;;; editorconfig-core-handle.el --- Handle Class for EditorConfig File  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2011-2019 EditorConfig Team
+;; Copyright (C) 2011-2021 EditorConfig Team
 
 ;; Author: EditorConfig Team <editorconfig@googlegroups.com>
 
@@ -67,10 +67,10 @@ IF not match, return nil."
 
 Slots:
 `top-props'
-  Alist of top propetties like ((\"root\" . \"true\"))
+  Alist of top properties like ((\"root\" . \"true\"))
 
 `sections'
-  List of `editorconfig-core-hadnle-section' strucure object.
+  List of `editorconfig-core-handle-section' structure objects.
 
 `mtime'
   Last modified time of .editorconfig file.
