@@ -1,11 +1,11 @@
 ;;; editorconfig.el --- EditorConfig Emacs Plugin  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2011-2019 EditorConfig Team
+;; Copyright (C) 2011-2021 EditorConfig Team
 
 ;; Author: EditorConfig Team <editorconfig@googlegroups.com>
-;; Version: 0.8.1
+;; Version: 0.8.2
 ;; URL: https://github.com/editorconfig/editorconfig-emacs#readme
-;; Package-Requires: ((cl-lib "0.5") (emacs "24"))
+;; Package-Requires: ((cl-lib "0.5") (nadvice "0.3") (emacs "24"))
 
 ;; See
 ;; https://github.com/editorconfig/editorconfig-emacs/graphs/contributors
@@ -39,6 +39,7 @@
 
 ;;; Code:
 (require 'cl-lib)
+(require 'nadvice)
 (eval-when-compile
   (require 'rx)
   (defvar tex-indent-basic)
@@ -48,7 +49,7 @@
 
 (declare-function editorconfig-core-get-properties-hash
                   "editorconfig-core"
-                  nil)
+                  (&optional file confname confversion))
 
 (defgroup editorconfig nil
   "EditorConfig Emacs Plugin.
@@ -59,26 +60,30 @@ coding styles between different editors and IDEs."
   :prefix "editorconfig-"
   :group 'tools)
 
-(defcustom editorconfig-exec-path
-  "editorconfig"
-  "Path to EditorConfig executable.
-
-Used by `editorconfig-call-editorconfig-exec'."
-  :type 'string
-  :group 'editorconfig)
 (define-obsolete-variable-alias
   'edconf-exec-path
   'editorconfig-exec-path
   "0.5")
+(defcustom editorconfig-exec-path
+  "editorconfig"
+  "Path to EditorConfig executable.
 
+Used by `editorconfig--execute-editorconfig-exec'."
+  :type 'string
+  :group 'editorconfig)
+
+(define-obsolete-variable-alias
+  'edconf-get-properties-function
+  'editorconfig-get-properties-function
+  "0.5")
 (defcustom editorconfig-get-properties-function
-  'editorconfig-get-properties
-  "A function which gets EditorConofig properties for current buffer.
+  'editorconfig-core-get-properties-hash
+  "A function which gets EditorConfig properties for specified file.
 
-This function will be called with no argument and should return a
-hash object containing properties, or nil if any core program is
-not available.  Keys of this hash should be symbols of properties, and values
-should be strings of their values.
+This function will be called with one argument, full path of the target file,
+and should return a hash object containing properties, or nil if any core
+program is not available.  Keys of this hash should be symbols of properties,
+and values should be strings of their values.
 
 
 For example, if you always want to use built-in core library instead
@@ -90,26 +95,30 @@ your init.el:
 
 Possible known values are:
 
-* `editorconfig-get-properties' (default)
+* `editorconfig-core-get-properties-hash' (default)
+  * Always use built-in Emacs-Lisp implementation to get properties
+* `editorconfig-get-properties'
   * Use `editorconfig-get-properties-from-exec' when
-    `editorconfig-exec-path' executable executable is found, otherwise
+    `editorconfig-exec-path' executable is found, otherwise
     use `editorconfig-core-get-properties-hash'
 * `editorconfig-get-properties-from-exec'
-  * Get properties by executing EditorConfig executable
-* `editorconfig-core-get-properties-hash'
-  * Always use built-in Emacs-Lisp implementation to get properties"
+  * Get properties by executing EditorConfig executable"
   :type 'function
   :group 'editorconfig)
-(define-obsolete-variable-alias
-  'edconf-get-properties-function
-  'editorconfig-get-properties-function
-  "0.5")
 
 (defcustom editorconfig-mode-lighter " EditorConfig"
-  "`editorconfig-mode' lighter string."
+  "Command `editorconfig-mode' lighter string."
   :type 'string
   :group 'editorconfig)
 
+(define-obsolete-variable-alias
+  'edconf-custom-hooks
+  'editorconfig-after-apply-functions
+  "0.5")
+(define-obsolete-variable-alias
+  'editorconfig-custom-hooks
+  'editorconfig-after-apply-functions
+  "0.7.14")
 (defcustom editorconfig-after-apply-functions ()
   "A list of functions after loading common EditorConfig settings.
 
@@ -126,17 +135,12 @@ show line numbers on the left:
     '(lambda (props)
        (let ((show-line-num (gethash 'emacs_linum props)))
          (cond ((equal show-line-num \"true\") (linum-mode 1))
-           ((equal show-line-num \"false\") (linum-mode 0))))))"
+           ((equal show-line-num \"false\") (linum-mode 0))))))
+
+This hook will be run even when there are no matching sections in
+\".editorconfig\", or no \".editorconfig\" file was found at all."
   :type 'hook
   :group 'editorconfig)
-(define-obsolete-variable-alias
-  'edconf-custom-hooks
-  'editorconfig-after-apply-functions
-  "0.5")
-(define-obsolete-variable-alias
-  'editorconfig-custom-hooks
-  'editorconfig-after-apply-functions
-  "0.7.14")
 
 (defcustom editorconfig-hack-properties-functions ()
   "A list of function to alter property values before applying them.
@@ -152,10 +156,17 @@ overwrite \"indent_style\" property when current `major-mode' is a
   (add-hook 'editorconfig-hack-properties-functions
             '(lambda (props)
                (when (derived-mode-p 'makefile-mode)
-                 (puthash 'indent_style \"tab\" props))))"
+                 (puthash 'indent_style \"tab\" props))))
+
+This hook will be run even when there are no matching sections in
+\".editorconfig\", or no \".editorconfig\" file was found at all."
   :type 'hook
   :group 'editorconfig)
 
+(define-obsolete-variable-alias
+  'edconf-indentation-alist
+  'editorconfig-indentation-alist
+  "0.5")
 (defcustom editorconfig-indentation-alist
   ;; For contributors: Sort modes in alphabetical order
   '((apache-mode apache-indent-level)
@@ -169,6 +180,7 @@ overwrite \"indent_style\" property when current `major-mode' is a
     (crystal-mode crystal-indent-level)
     (csharp-mode c-basic-offset)
     (css-mode css-indent-offset)
+    (d-mode c-basic-offset)
     (emacs-lisp-mode lisp-indent-offset)
     (enh-ruby-mode enh-ruby-indent-level)
     (erlang-mode erlang-indent-level)
@@ -211,6 +223,7 @@ overwrite \"indent_style\" property when current `major-mode' is a
     (livescript-mode livescript-tab-width)
     (lua-mode lua-indent-level)
     (matlab-mode matlab-indent-level)
+    (meson-mode meson-indent-basic)
     (mips-mode mips-tab-width)
     (mustache-mode mustache-basic-offset)
     (nasm-mode nasm-basic-offset)
@@ -229,13 +242,16 @@ overwrite \"indent_style\" property when current `major-mode' is a
     (pug-mode pug-tab-width)
     (puppet-mode puppet-indent-level)
     (python-mode . editorconfig-set-indentation-python-mode)
+    (rjsx-mode js-indent-level sgml-basic-offset)
     (ruby-mode ruby-indent-level)
     (rust-mode rust-indent-offset)
+    (rustic-mode rustic-indent-offset)
     (scala-mode scala-indent:step)
     (scss-mode css-indent-offset)
     (sgml-mode sgml-basic-offset)
     (sh-mode sh-basic-offset sh-indentation)
     (slim-mode slim-indent-offset)
+    (sml-mode sml-indent-level)
     (tcl-mode tcl-indent-level
               tcl-continued-indent-level)
     (terra-mode terra-indent-level)
@@ -286,25 +302,24 @@ NOTE: Only the **buffer local** value of VARIABLE will be set."
   :type '(alist :key-type symbol :value-type sexp)
   :risky t
   :group 'editorconfig)
-(define-obsolete-variable-alias
-  'edconf-indentation-alist
-  'editorconfig-indentation-alist
-  "0.5")
 
 (defcustom editorconfig-exclude-modes ()
   "Modes in which `editorconfig-mode-apply' will not run."
   :type '(repeat (symbol :tag "Major Mode"))
   :group 'editorconfig)
 
-(defcustom editorconfig-exclude-regexps
-  (list (eval-when-compile
-          (rx string-start (or "http" "https" "ftp" "sftp" "rsync") ":")))
+(defcustom editorconfig-exclude-regexps ()
   "List of regexp for buffer filenames `editorconfig-mode-apply' will not run.
 
 When variable `buffer-file-name' matches any of the regexps, then
 `editorconfig-mode-apply' will not do its work."
   :type '(repeat string)
   :group 'editorconfig)
+(with-eval-after-load 'recentf
+  (add-to-list 'editorconfig-exclude-regexps
+               (rx-to-string '(seq string-start
+                                   (eval (file-truename (expand-file-name recentf-save-file))))
+                             t)))
 
 (defcustom editorconfig-trim-whitespaces-mode nil
   "Buffer local minor-mode to use to trim trailing whitespaces.
@@ -324,7 +339,7 @@ current buffer yet.")
      t)
 
 (defvar editorconfig-lisp-use-default-indent nil
-  "Selectively ignore the value of indent_sizefor Lisp files.
+  "Selectively ignore the value of indent_size for Lisp files.
 Prevents `lisp-indent-offset' from being set selectively.
 
 nil - `lisp-indent-offset' is always set normally.
@@ -332,10 +347,29 @@ t   - `lisp-indent-offset' is never set normally
        (always use default indent for lisps).
 number - `lisp-indent-offset' is not set only if indent_size is
          equal to this number.  For example, if this is set to 2,
-         `lisp-indent-offset'will not be set only if indent_size is 2.")
+         `lisp-indent-offset' will not be set only if indent_size is 2.")
 
-(defconst editorconfig-unset-value "unset"
-  "String of value used to unset properties in .editorconfig .")
+(define-error 'editorconfig-error
+  "Error thrown from editorconfig lib")
+
+(defun editorconfig-error (&rest args)
+  "Signal an `editorconfig-error'.
+Make a message by passing ARGS to `format-message'."
+  (signal 'editorconfig-error (list (apply #'format-message args))))
+
+(defun editorconfig--disabled-for-filename (filename)
+  "Return non-nil when EditorConfig is disabled for FILENAME."
+  (cl-assert (stringp filename))
+  (cl-loop for regexp in editorconfig-exclude-regexps
+           if (string-match regexp filename) return t
+           finally return nil))
+
+(defun editorconfig--disabled-for-majormode (majormode)
+  "Return non-nil when Editorconfig is disabled for MAJORMODE."
+  (cl-assert majormode)
+  (or (editorconfig--provided-mode-derived-p majormode 'special-mode)
+      (memq majormode
+            editorconfig-exclude-modes)))
 
 (defun editorconfig-string-integer-p (string)
   "Return non-nil if STRING represents integer."
@@ -382,14 +416,13 @@ number - `lisp-indent-offset' is not set only if indent_size is
 
 (defun editorconfig-set-indentation (style &optional size tab_width)
   "Set indentation type from STYLE, SIZE and TAB_WIDTH."
-  (make-local-variable 'indent-tabs-mode)
-  (make-local-variable 'tab-width)
   (if (editorconfig-string-integer-p size)
       (setq size (string-to-number size))
     (when (not (equal size "tab")) (setq size nil)))
-  (setq tab-width (cond (tab_width (string-to-number tab_width))
-                        ((numberp size) size)
-                        (t tab-width)))
+  (cond (tab_width
+         (setq tab-width (string-to-number tab_width)))
+        ((numberp size)
+         (setq tab-width size)))
   (when (equal size "tab")
     (setq size tab-width))
   (cond ((equal style "space")
@@ -421,8 +454,15 @@ number - `lisp-indent-offset' is not set only if indent_size is
                                        ((integerp spec) (* spec size))
                                        (t spec))))))))))))))
 
-(defun editorconfig-set-coding-system (end-of-line charset)
-  "Set buffer coding system by END-OF-LINE and CHARSET."
+(defvar editorconfig--apply-coding-system-currently nil
+  "Used internally.")
+(make-variable-buffer-local 'editorconfig--apply-coding-system-currently)
+(put 'editorconfig--apply-coding-system-currently
+     'permanent-local
+     t)
+
+(defun editorconfig-merge-coding-systems (end-of-line charset)
+  "Return merged coding system symbol of END-OF-LINE and CHARSET."
   (let ((eol (cond
               ((equal end-of-line "lf") 'undecided-unix)
               ((equal end-of-line "cr") 'undecided-mac)
@@ -435,12 +475,44 @@ number - `lisp-indent-offset' is not set only if indent_size is
              ((equal charset "utf-16be") 'utf-16be-with-signature)
              ((equal charset "utf-16le") 'utf-16le-with-signature)
              (t 'undecided))))
-    (unless (and (eq eol 'undecided)
-                 (eq cs 'undecided))
-      (set-buffer-file-coding-system (merge-coding-systems
-                                      cs
-                                      eol)
-                                     nil t))))
+    (merge-coding-systems cs eol)))
+
+(cl-defun editorconfig-set-coding-system-revert (end-of-line charset)
+  "Set buffer coding system by END-OF-LINE and CHARSET.
+
+This function will revert buffer when the coding-system has been changed."
+  ;; `editorconfig--advice-find-file-noselect' does not use this function
+  (let ((coding-system (editorconfig-merge-coding-systems end-of-line
+                                                          charset)))
+    (display-warning '(editorconfig editorconfig-set-coding-system-revert)
+                     (format "editorconfig-set-coding-system-revert: buffer-file-name: %S | buffer-file-coding-system: %S | coding-system: %S | apply-currently: %S"
+                             buffer-file-name
+                             buffer-file-coding-system
+                             coding-system
+                             editorconfig--apply-coding-system-currently)
+                     :debug)
+    (when (eq coding-system 'undecided)
+      (cl-return-from editorconfig-set-coding-system-revert))
+    (when (and buffer-file-coding-system
+               (memq buffer-file-coding-system
+                     (coding-system-aliases (merge-coding-systems coding-system
+                                                                  buffer-file-coding-system))))
+      (cl-return-from editorconfig-set-coding-system-revert))
+    (unless (file-readable-p buffer-file-name)
+      (set-buffer-file-coding-system coding-system)
+      (cl-return-from editorconfig-set-coding-system-revert))
+    (unless (memq coding-system
+                  (coding-system-aliases editorconfig--apply-coding-system-currently))
+      ;; Revert functions might call editorconfig-apply again
+      (unwind-protect
+          (progn
+            (setq editorconfig--apply-coding-system-currently
+                  coding-system)
+            ;; Revert without query if buffer is not modified
+            (let ((revert-without-query '(".")))
+              (revert-buffer-with-coding-system coding-system)))
+        (setq editorconfig--apply-coding-system-currently
+              nil)))))
 
 (defun editorconfig-set-trailing-nl (final-newline)
   "Set up requiring final newline by FINAL-NEWLINE.
@@ -488,12 +560,6 @@ to non-nil when FINAL-NEWLINE is true."
              (> (string-to-number length) 0))
     (setq fill-column (string-to-number length))))
 
-(defvar editorconfig-file-type-emacs-whitelist
-  (append (mapcar 'car
-                  editorconfig-indentation-alist)
-          '(conf-mode))
-  "List of known `major-mode' that can be used for file_type_emacs value.")
-
 ;; Emacs<26 does not have provided-mode-derived-p
 (defun editorconfig--provided-mode-derived-p (mode &rest modes)
   "Non-nil if MODE is derived from one of MODES.
@@ -506,106 +572,41 @@ If you just want to check `major-mode', use `derived-mode-p'."
     mode))
 
 
-(defun editorconfig-set-major-mode-from-name (filetype)
-  "Set buffer `major-mode' by FILETYPE.
+(defun editorconfig--execute-editorconfig-exec (filename)
+  "Execute EditorConfig core with FILENAME and return output."
+  (if filename
+      (with-temp-buffer
+        (let ((remote (file-remote-p filename))
+              (remote-localname (file-remote-p filename
+                                               'localname)))
+          (display-warning '(editorconfig editorconfig--execute-editorconfig-exec)
+                           (format "editorconfig--execute-editorconfig-exec: filename: %S | remote: %S | remote-localname: %S"
+                                   filename
+                                   remote
+                                   remote-localname)
+                           :debug)
+          (if remote
+              (progn
+                (cd (concat remote "/"))
+                (setq filename remote-localname))
+            (cd "/")))
+        (display-warning '(editorconfig editorconfig--execute-editorconfig-exec)
+                         (format "editorconfig--execute-editorconfig-exec: default-directory: %S | filename: %S"
+                                 default-directory
+                                 filename
+                                 )
+                         :debug)
+        (if (eq 0
+                (process-file editorconfig-exec-path nil t nil filename))
+            (buffer-string)
+          (editorconfig-error (buffer-string))))
+    ""))
 
-FILETYPE should be s string like `\"ini\"`, if not nil or empty string."
-  (let ((mode (and filetype
-                   (not (string= filetype
-                                 ""))
-                   (intern (concat filetype
-                                   "-mode")))))
-    (when mode
-      (if (fboundp mode)
-          (if (apply 'editorconfig--provided-mode-derived-p mode
-                     editorconfig-file-type-emacs-whitelist)
-              (editorconfig-apply-major-mode-safely mode)
-            (display-warning :error (format "Major-mode `%S' is not listed in `%S'"
-                                            mode
-                                            'editorconfig-file-type-emacs-whitelist)))
-        (display-warning :error (format "Major-mode `%S' not found"
-                                        mode))
-        nil))))
-
-(defvar editorconfig--apply-major-mode-currently nil
-  "Used internally.")
-(make-variable-buffer-local 'editorconfig--apply-major-mode-currently)
-(put 'editorconfig--apply-major-mode-currently
-     'permanent-local
-     t)
-
-(defun editorconfig-apply-major-mode-safely (mode)
-  "Set `major-mode' to MODE.
-Normally `editorconfig-apply' will be hooked so that it runs when changing
-`major-mode', so there is a possibility that MODE is called infinitely if
-MODE is called naively from inside of `editorconfig-apply'.
-This funcion will avoid such cases and set `major-mode' safely.
-
-Just checking current `major-mode' value is not enough, because it can be
-different from MODE value (for example, `conf-mode' will set `major-mode' to
-`conf-unix-mode' or another conf mode)."
-  (cl-assert mode)
-  (when (and (not (eq mode
-                      editorconfig--apply-major-mode-currently))
-             (not (eq mode
-                      major-mode))
-             (not (derived-mode-p mode)))
-    (unwind-protect
-        (progn
-          (setq editorconfig--apply-major-mode-currently
-                mode)
-          (funcall mode))
-      (setq editorconfig--apply-major-mode-currently
-            nil))))
-
-(defun editorconfig--find-mode-from-ext (ext &optional filename)
-  "Get suitable `major-mode' from EXT and FILENAME.
-If FILENAME is omitted filename of current buffer is used."
-  (cl-assert ext)
-  (cl-assert (not (string= ext "")))
-  (let* ((name (concat (or filename
-                           buffer-file-name)
-                       "."
-                       ext)))
-    (assoc-default name
-                   auto-mode-alist
-                   'string-match)))
-
-(defun editorconfig-set-major-mode-from-ext (ext)
-  "Set buffer `major-mode' by EXT.
-
-EXT should be a string like `\"ini\"`, if not nil or empty string."
-  (cl-assert buffer-file-name)
-  (when (and ext
-             (not (string= ext ""))
-             (not (string= ext editorconfig-unset-value)))
-
-    (let ((mode (editorconfig--find-mode-from-ext ext
-                                                  buffer-file-name)))
-      (if mode
-          (editorconfig-apply-major-mode-safely mode)
-        (display-warning :error (format "Major-mode for `%s' not found"
-                                        ext))
-        nil))))
-
-(defun editorconfig-call-editorconfig-exec ()
-  "Call EditorConfig core and return output."
-  (let* ((filename (buffer-file-name))
-         (filename (and filename (expand-file-name filename))))
-    (if filename
-        (with-temp-buffer
-          (setq default-directory "/")
-          (if (eq 0
-                  (call-process editorconfig-exec-path nil t nil filename))
-              (buffer-string)
-            (error (buffer-string))))
-      "")))
-
-(defun editorconfig-parse-properties (props-string)
+(defun editorconfig--parse-properties (props-string)
   "Create properties hash table from PROPS-STRING."
   (let (props-list properties)
     (setq props-list (split-string props-string "\n")
-          properties (make-hash-table :test 'equal))
+          properties (make-hash-table))
     (dolist (prop props-list properties)
       (let ((key-val (split-string prop " *= *")))
         (when (> (length key-val) 1)
@@ -613,85 +614,240 @@ EXT should be a string like `\"ini\"`, if not nil or empty string."
                 (val (mapconcat 'identity (cdr key-val) "")))
             (puthash key val properties)))))))
 
-(defun editorconfig-get-properties-from-exec ()
-  "Get EditorConfig properties of current buffer.
+(defun editorconfig-get-properties-from-exec (filename)
+  "Get EditorConfig properties of file FILENAME.
 
 This function uses value of `editorconfig-exec-path' to get properties."
   (if (executable-find editorconfig-exec-path)
-      (editorconfig-parse-properties (editorconfig-call-editorconfig-exec))
-    (error "Unable to find editorconfig executable")))
+      (editorconfig--parse-properties (editorconfig--execute-editorconfig-exec filename))
+    (editorconfig-error "Unable to find editorconfig executable")))
 
-(defun editorconfig-get-properties ()
-  "Get EditorConfig properties of current buffer.
+(defun editorconfig-get-properties (filename)
+  "Get EditorConfig properties for file FILENAME.
 
 It calls `editorconfig-get-properties-from-exec' if
 `editorconfig-exec-path' is found, otherwise
 `editorconfig-core-get-properties-hash'."
   (if (and (executable-find editorconfig-exec-path)
            (not (file-remote-p buffer-file-name)))
-      (editorconfig-get-properties-from-exec)
+      (editorconfig-get-properties-from-exec filename)
     (require 'editorconfig-core)
-    (editorconfig-core-get-properties-hash)))
+    (editorconfig-core-get-properties-hash filename)))
+
+(defun editorconfig-call-get-properties-function (filename)
+  "Call `editorconfig-get-properties-function' with FILENAME and return result.
+
+This function also removes 'unset'ted properties and calls
+`editorconfig-hack-properties-functions'."
+  (unless (functionp editorconfig-get-properties-function)
+    (editorconfig-error "Invalid editorconfig-get-properties-function value"))
+  (if (stringp filename)
+      (setq filename (expand-file-name filename))
+    (editorconfig-error "Invalid argument: %S" filename))
+  (let ((props nil))
+    (condition-case err
+        (setq props (funcall editorconfig-get-properties-function
+                             filename))
+      (error
+       (editorconfig-error "Error from editorconfig-get-properties-function: %S"
+                           err)))
+    (cl-loop for k being the hash-keys of props using (hash-values v)
+             when (equal v "unset") do (remhash k props))
+    props))
+
+(defun editorconfig-set-local-variables (props)
+  "Set buffer variables according to EditorConfig PROPS."
+  (editorconfig-set-indentation (gethash 'indent_style props)
+                                (gethash 'indent_size props)
+                                (gethash 'tab_width props))
+  (editorconfig-set-trailing-nl (gethash 'insert_final_newline props))
+  (editorconfig-set-trailing-ws (gethash 'trim_trailing_whitespace props))
+  (editorconfig-set-line-length (gethash 'max_line_length props)))
 
 ;;;###autoload
 (defun editorconfig-apply ()
   "Get and apply EditorConfig properties to current buffer.
-This function ignores `editorconfig-exclude-modes' and
-`editorconfig-exclude-regexps', and always applies available properties."
+
+This function does not respect the values of `editorconfig-exclude-modes' and
+`editorconfig-exclude-regexps' and always applies available properties.
+Use `editorconfig-mode-apply' instead to make use of these variables."
   (interactive)
   (when buffer-file-name
     (condition-case err
         (progn
-          (unless (functionp editorconfig-get-properties-function)
-            (error "Invalid editorconfig-get-properties-function value"))
-          (let ((props (funcall editorconfig-get-properties-function)))
-            (progn
-              (condition-case err
-                  (run-hook-with-args 'editorconfig-hack-properties-functions props)
-                (error
-                 (display-warning 'editorconfig-hack-properties-functions
-                                  (concat (error-message-string err)
-                                          ". Abort running hook.")
-                                  :warning)))
-              (setq editorconfig-properties-hash props)
-              (editorconfig-set-indentation (gethash 'indent_style props)
-                                            (gethash 'indent_size props)
-                                            (gethash 'tab_width props))
-              (editorconfig-set-coding-system
-               (gethash 'end_of_line props)
-               (gethash 'charset props))
-              (editorconfig-set-trailing-nl (gethash 'insert_final_newline props))
-              (editorconfig-set-trailing-ws (gethash 'trim_trailing_whitespace props))
-              (editorconfig-set-line-length (gethash 'max_line_length props))
-              (editorconfig-set-major-mode-from-name (gethash 'file_type_emacs props))
-              (editorconfig-set-major-mode-from-ext (gethash 'file_type_ext props))
-              (condition-case err
-                  (run-hook-with-args 'editorconfig-after-apply-functions props)
-                (error
-                 (display-warning 'editorconfig-after-apply-functions
-                                  (concat (error-message-string err)
-                                          ". Stop running hook.")
-                                  :warning))))))
+          (let ((props (editorconfig-call-get-properties-function buffer-file-name)))
+            (condition-case err
+                (run-hook-with-args 'editorconfig-hack-properties-functions props)
+              (error
+               (display-warning '(editorconfig editorconfig-hack-properties-functions)
+                                (format "Error while running editorconfig-hack-properties-functions, abort running hook: %S"
+                                        err)
+                                :warning)))
+            (setq editorconfig-properties-hash props)
+            (editorconfig-set-local-variables props)
+            (editorconfig-set-coding-system-revert
+             (gethash 'end_of_line props)
+             (gethash 'charset props))
+            (condition-case err
+                (run-hook-with-args 'editorconfig-after-apply-functions props)
+              (error
+               (display-warning '(editorconfig editorconfig-after-apply-functions)
+                                (format "Error while running editorconfig-after-apply-functions, abort running hook: %S"
+                                        err)
+                                :warning)))))
       (error
-       (display-warning 'editorconfig
-                        (concat (error-message-string err)
-                                ".  Styles will not be applied.")
+       (display-warning '(editorconfig editorconfig-apply)
+                        (format "Error in editorconfig-apply, styles will not be applied: %S" err)
                         :error)))))
 
 (defun editorconfig-mode-apply ()
   "Get and apply EditorConfig properties to current buffer.
+
 This function does nothing when the major mode is listed in
 `editorconfig-exclude-modes', or variable `buffer-file-name' matches
 any of regexps in `editorconfig-exclude-regexps'."
+  (interactive)
   (when (and major-mode
-             (not (memq major-mode
-                        editorconfig-exclude-modes))
+             (not (editorconfig--disabled-for-majormode major-mode))
              buffer-file-name
-             (not (cl-loop for regexp in editorconfig-exclude-regexps
-                           if (string-match regexp buffer-file-name) return t
-                           finally return nil)))
+             (not (editorconfig--disabled-for-filename buffer-file-name)))
     (editorconfig-apply)))
 
+(defun editorconfig-major-mode-hook ()
+  "Function to run when `major-mode' has been changed.
+
+This functions does not reload .editorconfig file, just sets local variables
+again.  Changing major mode can reset these variables.
+
+This function also executes `editorconfig-after-apply-functions' functions."
+  (display-warning '(editorconfig editorconfig-major-mode-hook)
+                   (format "editorconfig-major-mode-hook: editorconfig-mode: %S, major-mode: %S, -properties-hash: %S"
+                           (and (boundp 'editorconfig-mode)
+                                editorconfig-mode)
+                           major-mode
+                           editorconfig-properties-hash)
+                   :debug)
+  (when (and (boundp 'editorconfig-mode)
+             editorconfig-mode
+             editorconfig-properties-hash)
+    (editorconfig-set-local-variables editorconfig-properties-hash)
+    (condition-case err
+        (run-hook-with-args 'editorconfig-after-apply-functions editorconfig-properties-hash)
+      (error
+       (display-warning '(editorconfig editorconfig-major-mode-hook)
+                        (format "Error while running `editorconfig-after-apply-functions': %S"
+                                err))))))
+
+(defvar editorconfig--cons-filename-codingsystem nil
+  "Used interally.
+
+`editorconfig--advice-find-file-noselect' will set this variable, and
+`editorconfig--advice-insert-file-contents' will use this variable to set
+`coding-system-for-read' value.")
+
+(defun editorconfig--advice-insert-file-contents (f filename &rest args)
+  "Set `coding-system-for-read'.
+
+This function should be added as an advice function to `insert-file-contents'.
+F is that function, and FILENAME and ARGS are arguments passed to F."
+  ;; This function uses `editorconfig--cons-filename-codingsystem' to decide what coding-system
+  ;; should be used, which will be set by `editorconfig--advice-find-file-noselect'.
+  (display-warning '(editorconfig editorconfig--advice-insert-file-contents)
+                   (format "editorconfig--advice-insert-file-contents: filename: %S args: %S codingsystem: %S bufferfilename: %S"
+                           filename args
+                           editorconfig--cons-filename-codingsystem
+                           buffer-file-name)
+                   :debug)
+  (if (and (stringp filename)
+           (stringp (car editorconfig--cons-filename-codingsystem))
+           (string= (expand-file-name filename)
+                    (car editorconfig--cons-filename-codingsystem))
+           (cdr editorconfig--cons-filename-codingsystem)
+           (not (eq (cdr editorconfig--cons-filename-codingsystem)
+                    'undecided)))
+      (let (
+            (coding-system-for-read (cdr editorconfig--cons-filename-codingsystem))
+            ;; (coding-system-for-read 'undecided)
+            )
+        (apply f filename args))
+    (apply f filename args)))
+
+(defun editorconfig--advice-find-file-noselect (f filename &rest args)
+  "Get EditorConfig properties and apply them to buffer to be visited.
+
+This function should be added as an advice function to `find-file-noselect'.
+F is that function, and FILENAME and ARGS are arguments passed to F."
+  (let ((props nil)
+        (coding-system nil)
+        (ret nil))
+    (condition-case err
+        (when (and (stringp filename)
+                   (not (editorconfig--disabled-for-filename filename)))
+          (setq props (editorconfig-call-get-properties-function filename))
+          (setq coding-system
+                (editorconfig-merge-coding-systems (gethash 'end_of_line props)
+                                                   (gethash 'charset props))))
+      (error
+       (display-warning '(editorconfig editorconfig--advice-find-file-noselect)
+                        (format "Failed to get properties, styles will not be applied: %S"
+                                err)
+                        :warning)))
+
+    (let ((editorconfig--cons-filename-codingsystem (cons (expand-file-name filename)
+                                                          coding-system)))
+      (setq ret (apply f filename args)))
+
+    (condition-case err
+        (with-current-buffer ret
+          (when (and props
+                     ;; filename has already been checked
+                     (not (editorconfig--disabled-for-majormode major-mode)))
+            (when (and (file-remote-p filename)
+                       (not (local-variable-p 'buffer-file-coding-system))
+                       (not (file-exists-p filename))
+                       coding-system
+                       (not (eq coding-system
+                                'undecided)))
+              ;; When file path indicates it is a remote file and it actually
+              ;; does not exists, `buffer-file-coding-system' will not be set.
+              ;; (Seems `insert-file-contents' will not be called)
+              ;; For that case, explicitly set this value so that saving will be done
+              ;; with expected coding system.
+              (set-buffer-file-coding-system coding-system))
+
+            ;; NOTE: When using editorconfig-2-mode, hack-properties-functions cannot affect coding-system value,
+            ;; because it has to be set before initializing buffers.
+            (condition-case err
+                (run-hook-with-args 'editorconfig-hack-properties-functions props)
+              (error
+               (display-warning '(editorconfig editorconfig-hack-properties-functions)
+                                (format "Error while running editorconfig-hack-properties-functions, abort running hook: %S"
+                                        err)
+                                :warning)))
+            (setq editorconfig-properties-hash props)
+            ;; When initializing buffer, `editorconfig-major-mode-hook'
+            ;; will be called before setting `editorconfig-properties-hash', so
+            ;; execute this explicitly here.
+            (editorconfig-set-local-variables props)
+
+
+            (condition-case err
+                (run-hook-with-args 'editorconfig-after-apply-functions props)
+              (error
+               (display-warning '(editorconfig editorconfig--advice-find-file-noselect)
+                                (format "Error while running `editorconfig-after-apply-functions': %S"
+                                        err))))))
+      (error
+       (display-warning '(editorconfig editorconfig--advice-find-file-noselect)
+                        (format "Error while setting variables from EditorConfig: %S" err))))
+    ret))
+
+(defvar editorconfig--enable-20210221-testing nil
+  "Enable testing version of `editorconfig-mode'.
+
+Currently this mode is not well tested yet and can cause unexpected behaviors
+like killing Emacs process or not able to visit files at all.")
+
 ;;;###autoload
 (define-minor-mode editorconfig-mode
   "Toggle EditorConfig feature.
@@ -700,19 +856,44 @@ To disable EditorConfig in some buffers, modify
 `editorconfig-exclude-modes' or `editorconfig-exclude-regexps'."
   :global t
   :lighter editorconfig-mode-lighter
-  ;; See https://github.com/editorconfig/editorconfig-emacs/issues/141 for why
-  ;; not `after-change-major-mode-hook'
-  (dolist (hook '(change-major-mode-after-body-hook
-                  read-only-mode-hook
-                  ;; Some modes call `kill-all-local-variables' in their init
-                  ;; code, which clears some values set by editorconfig.
-                  ;; For those modes, editorconfig-apply need to be called
-                  ;; explicitly through their hooks.
-                  rpm-spec-mode-hook
-                  ))
-    (if editorconfig-mode
-        (add-hook hook 'editorconfig-mode-apply)
-      (remove-hook hook 'editorconfig-mode-apply))))
+  (if editorconfig--enable-20210221-testing
+      (let ((modehooks '(prog-mode-hook
+                          text-mode-hook
+                          read-only-mode-hook
+                          ;; Some modes call `kill-all-local-variables' in their init
+                          ;; code, which clears some values set by editorconfig.
+                          ;; For those modes, editorconfig-apply need to be called
+                          ;; explicitly through their hooks.
+                          rpm-spec-mode-hook
+                          )))
+        (if editorconfig-mode
+            (progn
+              (advice-add 'find-file-noselect :around 'editorconfig--advice-find-file-noselect)
+              (advice-add 'insert-file-contents :around 'editorconfig--advice-insert-file-contents)
+              (dolist (hook modehooks)
+                (add-hook hook
+                          'editorconfig-major-mode-hook
+                          t)))
+          (advice-remove 'find-file-noselect 'editorconfig--advice-find-file-noselect)
+          (advice-remove 'insert-file-contents 'editorconfig--advice-insert-file-contents)
+          (dolist (hook modehooks)
+            (remove-hook hook
+                         'editorconfig-major-mode-hook))))
+
+    ;; editorconfig--enable-20210221-testing is disabled
+    ;; See https://github.com/editorconfig/editorconfig-emacs/issues/141 for why
+    ;; not `after-change-major-mode-hook'
+    (dolist (hook '(change-major-mode-after-body-hook
+                    read-only-mode-hook
+                    ;; Some modes call `kill-all-local-variables' in their init
+                    ;; code, which clears some values set by editorconfig.
+                    ;; For those modes, editorconfig-apply need to be called
+                    ;; explicitly through their hooks.
+                    rpm-spec-mode-hook
+                    ))
+      (if editorconfig-mode
+          (add-hook hook 'editorconfig-mode-apply)
+        (remove-hook hook 'editorconfig-mode-apply)))))
 
 
 ;; Tools
@@ -773,7 +954,7 @@ To disable EditorConfig in some buffers, modify
 (declare-function lm-version "lisp-mnt" nil)
 
 ;;;###autoload
-(defun  editorconfig-version (&optional show-version)
+(defun editorconfig-version (&optional show-version)
   "Get EditorConfig version as string.
 
 If called interactively or if SHOW-VERSION is non-nil, show the
@@ -786,7 +967,7 @@ version in the echo area and the messages buffer."
             (require 'lisp-mnt)
             (lm-version)))
          (pkg
-          (and (require 'package nil t)
+          (and (eval-and-compile (require 'package nil t))
                (cadr (assq 'editorconfig
                            package-alist))))
          (pkg-version
